defmodule MyResolver do

  @max_depth 6 # Remember the many denial-of-service attacks using various tricks to exhaust the resolver.
  @root "193.0.14.129"  # K-root server. A real resolver would do priming (RFC 8109)
  @verbose true
  
  def usage() do
    IO.puts("Usage: myresolver name [type in lowercase]")
  end
  
  def find_a_server_with_ip_address(nslist) do
    if length(nslist) == 0 do
      {:error, :empty}
    else
      server = elem(Enum.fetch!(nslist, 0), 6) # We are lazy, we use
      # only the first server of the list. A real resolver would
      # choose at random, then memorize the fastest one, and be able
      # to switch to another one if there is a timeout.
      query(server, :a, @root, 0)
    end
  end
  
  def query(name, type \\ :aaaa, server \\ @root, depth \\ 0) do
    if depth > @max_depth do 
      {:error, :recursion_too_deep}
    else
      server_str = if not is_tuple(server) do
	server
      else
	to_string(:inet_parse.ntoa(server))
      end	
      if @verbose do
	IO.puts("Querying #{server_str} at depth #{depth}")
      end
      # DNS.resolve is to talk to resolvers. We talk to authoritative servers only.
      result = DNS.query(name, type, {server, 53})
      # The terminology of the library is strange. nslist is the
      # Authority section, anlist the Additional, anlist the Answer.
      case result.header.rcode do
	0 -> # Do we have a final result of the requested type? ("Best
	     # case" in J. Evans' program)
	  case Enum.fetch(Enum.filter(result.anlist,
		    fn x -> x.type == type end), 0) do
	    {:ok, data} -> {:ok, data.data}
	    :error -> # No. Do we have a delegation?
	      if length(Enum.filter(result.nslist, fn x -> elem(x, 2) == :ns end)) != 0 do
		# We filter only IPv4 IP addresses since the DNS
		# library does not support IPv6
		# https://github.com/tungd/elixir-dns/issues/43
		new_server =
		  case Enum.fetch(Enum.filter(result.arlist, fn x -> x.type == :a end), 0) do
		    {:ok, address} -> address.data # Use the glue, Luke ("Second best" in
                               		           # J. Evan's program)
		    :error -> # No glue, we have to do an explicit resolution. ("Third best"
                              # in J. Evan's program)
		      case find_a_server_with_ip_address(result.nslist) do
			{:ok, address} -> address
			other -> other
		      end
		  end
		# Let's recurse
		query(name,
		  type,
		  new_server,
		  depth+1)
	      else # No error, no answer of the proper type, no
		   # delegation, this is NODATA (J. Evans' program
		   # panics here)
		{:error, :nodata}
	      end
	  end
	3 -> {:error, :nxdomain}
	other -> {:error, other} # A real resolver would try another authoritative name server
      end
    end    
  end
  
end

{name, type} = case length(System.argv()) do
  2 ->
    {Enum.fetch!(System.argv(), 0), String.to_atom(Enum.fetch!(System.argv(), 1))}
  1 ->
    {Enum.fetch!(System.argv(), 0), :aaaa}
  _ ->
    MyResolver.usage()
    exit("Wrong usage")
end	    
case MyResolver.query(name, type) do
  {:ok, value} -> IO.puts("Value for #{name}/#{type} is #{inspect value}")
  {:error, reason} -> IO.puts("Cannot get #{name}/#{type}: #{reason}")
end

